import inspect

from dotenv import load_dotenv

from lion.integrations.openai_.api_endpoints.api_request import OpenAIRequest
from lion.integrations.openai_.api_endpoints.match_data_model import match_data_model
from lion.service import Service, register_service

from .OpenAIModel import OpenAIModel

load_dotenv()


@register_service
class OpenAIService(Service):
    def __init__(
        self,
        api_key: str,
        openai_organization: str = None,
        openai_project: str = None,
        name: str = None,
    ):
        super().__setattr__("_initialized", False)
        self.api_key = api_key
        self.openai_organization = openai_organization
        self.openai_project = openai_project
        self.name = name
        self.rate_limiters = {}  # model: RateLimiter
        super().__setattr__("_initialized", True)

    def __setattr__(self, key, value):
        if getattr(self, "_initialized", False) and key in [
            "api_key",
            "openai_organization",
            "openai_project",
        ]:
            raise AttributeError(
                f"Cannot modify '{key}' after initialization. "
                f"Please set a new service object for new keys."
            )
        super().__setattr__(key, value)

    def check_rate_limiter(
        self,
        openai_model: OpenAIModel,
        limit_requests: int = None,
        limit_tokens: int = None,
    ):
        shared_models = {
            "gpt-4-turbo-2024-04-09": "gpt-4-turbo",
            "gpt-4-turbo-preview": "gpt-4-turbo",
            "gpt-4-0125-preview": "gpt-4-turbo",
            "gpt-4-1106-preview": "gpt-4-turbo",
            "gpt-4o-2024-05-13": "gpt-4o",
            "gpt-4o-2024-08-06": "gpt-4o",
            "gpt-4o-mini-2024-07-18": "gpt-4o-mini",
        }

        if openai_model.model in shared_models:
            model = shared_models[openai_model.model]
        else:
            model = openai_model.model

        if model not in self.rate_limiters:
            self.rate_limiters[model] = openai_model.rate_limiter
        else:
            openai_model.rate_limiter = self.rate_limiters[model]
            if limit_requests:
                openai_model.rate_limiter.limit_requests = limit_requests
            if limit_tokens:
                openai_model.rate_limiter.limit_tokens = limit_tokens

        return openai_model

    @staticmethod
    def match_data_model(task_name):
        return match_data_model(task_name)

    @classmethod
    def list_tasks(cls):
        methods = []
        for name, member in inspect.getmembers(cls, predicate=inspect.isfunction):
            if name not in [
                "__init__",
                "__setattr__",
                "check_rate_limiter",
                "match_data_model",
            ]:
                methods.append(name)
        return methods

    # Audio
    def create_speech(self, model: str, limit_requests: int = None):
        model_obj = OpenAIModel(
            model=model,
            api_key=self.api_key,
            openai_organization=self.openai_organization,
            openai_project=self.openai_project,
            endpoint="audio/speech",
            method="POST",
            content_type="application/json",
            limit_requests=limit_requests,
        )

        return self.check_rate_limiter(model_obj, limit_requests=limit_requests)

    def create_transcription(self, model: str, limit_requests: int = None):
        model_obj = OpenAIModel(
            model=model,
            api_key=self.api_key,
            openai_organization=self.openai_organization,
            openai_project=self.openai_project,
            endpoint="audio/transcriptions",
            method="POST",
            limit_requests=limit_requests,
        )

        return self.check_rate_limiter(model_obj, limit_requests=limit_requests)

    def create_translation(self, model: str, limit_requests: int = None):
        model_obj = OpenAIModel(
            model=model,
            api_key=self.api_key,
            openai_organization=self.openai_organization,
            openai_project=self.openai_project,
            endpoint="audio/translations",
            method="POST",
            limit_requests=limit_requests,
        )

        return self.check_rate_limiter(model_obj, limit_requests=limit_requests)

    # Chat
    def create_chat_completion(
        self, model: str, limit_tokens: int = None, limit_requests: int = None
    ):
        model_obj = OpenAIModel(
            model=model,
            api_key=self.api_key,
            openai_organization=self.openai_organization,
            openai_project=self.openai_project,
            endpoint="chat/completions",
            method="POST",
            content_type="application/json",
            limit_tokens=limit_tokens,
            limit_requests=limit_requests,
        )

        return self.check_rate_limiter(
            model_obj, limit_requests=limit_requests, limit_tokens=limit_tokens
        )

    # Embeddings
    def create_embeddings(
        self, model: str, limit_tokens: int = None, limit_requests: int = None
    ):
        model_obj = OpenAIModel(
            model=model,
            api_key=self.api_key,
            openai_organization=self.openai_organization,
            openai_project=self.openai_project,
            endpoint="embeddings",
            method="POST",
            content_type="application/json",
            limit_tokens=limit_tokens,
            limit_requests=limit_requests,
        )
        return self.check_rate_limiter(
            model_obj, limit_requests=limit_requests, limit_tokens=limit_tokens
        )

    # Fine_tuning
    def create_fine_tuning_job(self):
        return OpenAIRequest(
            api_key=self.api_key,
            openai_organization=self.openai_organization,
            openai_project=self.openai_project,
            endpoint="fine_tuning/jobs",
            method="POST",
            content_type="application/json",
        )

    def list_fine_tuning_jobs(self):
        return OpenAIRequest(
            api_key=self.api_key,
            openai_organization=self.openai_organization,
            openai_project=self.openai_project,
            endpoint="fine_tuning/jobs",
            method="GET",
        )

    def list_fine_tuning_events(self):
        return OpenAIRequest(
            api_key=self.api_key,
            openai_organization=self.openai_organization,
            openai_project=self.openai_project,
            endpoint="fine_tuning/jobs/{fine_tuning_job_id}/events",
            method="GET",
        )

    def list_fine_tuning_checkpoints(self):
        return OpenAIRequest(
            api_key=self.api_key,
            openai_organization=self.openai_organization,
            openai_project=self.openai_project,
            endpoint="fine_tuning/jobs/{fine_tuning_job_id}/checkpoints",
            method="GET",
        )

    def retrieve_fine_tuning_job(self):
        return OpenAIRequest(
            api_key=self.api_key,
            openai_organization=self.openai_organization,
            openai_project=self.openai_project,
            endpoint="fine_tuning/jobs/{fine_tuning_job_id}",
            method="GET",
        )

    def cancel_fine_tuning(self):
        return OpenAIRequest(
            api_key=self.api_key,
            openai_organization=self.openai_organization,
            openai_project=self.openai_project,
            endpoint="fine_tuning/jobs/{fine_tuning_job_id}/cancel",
            method="POST",
        )

    # Batch
    def create_batch(self):
        return OpenAIRequest(
            api_key=self.api_key,
            openai_organization=self.openai_organization,
            openai_project=self.openai_project,
            endpoint="batches",
            method="POST",
            content_type="application/json",
        )

    def retrieve_batch(self):
        return OpenAIRequest(
            api_key=self.api_key,
            openai_organization=self.openai_organization,
            openai_project=self.openai_project,
            endpoint="batches/{batch_id}",
            method="GET",
            content_type="application/json",
        )

    def cancel_batch(self):
        return OpenAIRequest(
            api_key=self.api_key,
            openai_organization=self.openai_organization,
            openai_project=self.openai_project,
            endpoint="batches/{batch_id}/cancel",
            method="POST",
            content_type="application/json",
        )

    def list_batch(self):
        return OpenAIRequest(
            api_key=self.api_key,
            openai_organization=self.openai_organization,
            openai_project=self.openai_project,
            endpoint="batches",
            method="GET",
            content_type="application/json",
        )

    # Files
    def upload_file(self):
        return OpenAIRequest(
            api_key=self.api_key,
            openai_organization=self.openai_organization,
            openai_project=self.openai_project,
            endpoint="files",
            method="POST",
        )

    def list_files(self):
        return OpenAIRequest(
            api_key=self.api_key,
            openai_organization=self.openai_organization,
            openai_project=self.openai_project,
            endpoint="files",
            method="GET",
        )

    def retrieve_file(self):
        return OpenAIRequest(
            api_key=self.api_key,
            openai_organization=self.openai_organization,
            openai_project=self.openai_project,
            endpoint="files/{file_id}",
            method="GET",
        )

    def delete_file(self):
        return OpenAIRequest(
            api_key=self.api_key,
            openai_organization=self.openai_organization,
            openai_project=self.openai_project,
            endpoint="files/{file_id}",
            method="DELETE",
        )

    def retrieve_file_content(self):
        return OpenAIRequest(
            api_key=self.api_key,
            openai_organization=self.openai_organization,
            openai_project=self.openai_project,
            endpoint="files/{file_id}/content",
            method="GET",
        )

    # Uploads
    def create_upload(self):
        return OpenAIRequest(
            api_key=self.api_key,
            openai_organization=self.openai_organization,
            openai_project=self.openai_project,
            endpoint="uploads",
            method="POST",
        )

    def add_upload_part(self):
        return OpenAIRequest(
            api_key=self.api_key,
            openai_organization=self.openai_organization,
            openai_project=self.openai_project,
            endpoint="uploads/{upload_id}/parts",
            method="POST",
        )

    def complete_upload(self):
        return OpenAIRequest(
            api_key=self.api_key,
            openai_organization=self.openai_organization,
            openai_project=self.openai_project,
            endpoint="uploads/{upload_id}/complete",
            method="POST",
        )

    def cancel_upload(self):
        return OpenAIRequest(
            api_key=self.api_key,
            openai_organization=self.openai_organization,
            openai_project=self.openai_project,
            endpoint="uploads/{upload_id}/cancel",
            method="POST",
        )

    # Images
    def create_image(self):
        return OpenAIRequest(
            api_key=self.api_key,
            openai_organization=self.openai_organization,
            openai_project=self.openai_project,
            endpoint="images/generations",
            method="POST",
            content_type="application/json",
        )

    def create_image_edit(self):
        return OpenAIRequest(
            api_key=self.api_key,
            openai_organization=self.openai_organization,
            openai_project=self.openai_project,
            endpoint="images/edits",
            method="POST",
        )

    def create_image_variation(self):
        return OpenAIRequest(
            api_key=self.api_key,
            openai_organization=self.openai_organization,
            openai_project=self.openai_project,
            endpoint="images/variations",
            method="POST",
        )

    # Models
    def list_models(self):
        return OpenAIRequest(
            api_key=self.api_key,
            openai_organization=self.openai_organization,
            openai_project=self.openai_project,
            endpoint="models",
            method="GET",
        )

    def retrieve_model(self):
        return OpenAIRequest(
            api_key=self.api_key,
            openai_organization=self.openai_organization,
            openai_project=self.openai_project,
            endpoint="models/{model}",
            method="GET",
        )

    def delete_fine_tuned_model(self):
        return OpenAIRequest(
            api_key=self.api_key,
            openai_organization=self.openai_organization,
            openai_project=self.openai_project,
            endpoint="models/{model}",
            method="DELETE",
        )

    # Moderations
    def create_moderation(self):
        return OpenAIRequest(
            api_key=self.api_key,
            openai_organization=self.openai_organization,
            openai_project=self.openai_project,
            endpoint="moderations",
            method="POST",
            content_type="application/json",
        )



from pydantic import BaseModel, ConfigDict


class AnthropicEndpointRequestBody(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
        use_enum_values=True,
        validate_assignment=True,
        arbitrary_types_allowed=True,
    )


class AnthropicEndpointResponseBody(BaseModel):
    model_config = ConfigDict(
        use_enum_values=True,
        validate_assignment=True,
        arbitrary_types_allowed=True,
    )


class AnthropicEndpointQueryParam(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
        use_enum_values=True,
        validate_assignment=True,
        arbitrary_types_allowed=True,
    )


class AnthropicEndpointPathParam(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
        use_enum_values=True,
        validate_assignment=True,
        arbitrary_types_allowed=True,
    )



from pydantic import BaseModel, ConfigDict, Field

from .log_prob_models import LogProbs
from .message_models import Message
from .types import FinishReason


class Choice(BaseModel):
    finish_reason: FinishReason = Field(
        description=(
            "The reason the model stopped generating tokens. This will be "
            "stop if the model hit a natural stop point or a provided stop "
            "sequence, length if the maximum number of tokens specified in "
            "the request was reached, content_filter if content was omitted "
            "due to a flag from our content filters, tool_calls if the model "
            "called a tool, or function_call (deprecated) if the model "
            "called a function."
        )
    )

    index: int = Field(description="The index of the choice in the list of choices.")

    message: Message = Field(
        description="A chat completion message generated by the model."
    )

    logprobs: LogProbs | None = Field(
        None, description="Log probability information for the choice."
    )

    model_config = ConfigDict(use_enum_values=True)


class ChunkChoice(BaseModel):
    delta: Message = Field(
        description="A chat completion delta generated" " by streamed model responses."
    )

    logprobs: LogProbs | None = Field(
        description="Log probability information for the choice."
    )

    finish_reason: FinishReason | None = Field(
        description=(
            "The reason the model stopped generating tokens. This will be "
            "stop if the model hit a natural stop point or a provided stop "
            "sequence, length if the maximum number of tokens specified in "
            "the request was reached, content_filter if content was omitted "
            "due to a flag from our content filters, tool_calls if the model "
            "called a tool, or function_call (deprecated) if the model "
            "called a function."
        )
    )

    index: int = Field(description="The index of the choice in the list of choices.")

    model_config = ConfigDict(use_enum_values=True)



from pydantic import BaseModel, field_validator

from lion.core.models import FieldModel
from lion.libs.parse import to_num

from .prompts import confidence_description


def validate_confidence_score(cls, value) -> float:
    try:
        return to_num(
            value,
            upper_bound=1,
            lower_bound=0,
            num_type=float,
            precision=3,
        )
    except Exception:
        return -1


CONFIDENCE_SCORE_FIELD = FieldModel(
    name="confidence_score",
    annotation=float | None,
    default=None,
    title="Confidence Score",
    description=confidence_description,
    examples=[0.821, 0.257, 0.923, 0.439],
    validator=validate_confidence_score,
    validator_kwargs={"mode": "before"},
)


class ReasonModel(BaseModel):

    title: str | None = None
    content: str | None = None
    confidence_score: float | None = CONFIDENCE_SCORE_FIELD.field_info

    @field_validator("confidence_score", **CONFIDENCE_SCORE_FIELD.validator_kwargs)
    def _validate_confidence(cls, v):
        return CONFIDENCE_SCORE_FIELD.validator(cls, v)


REASON_FIELD = FieldModel(
    name="reason",
    annotation=ReasonModel | None,
    default=None,
    title="Reason",
    description="**Provide a concise reason for the decision made.**",
)


__all__ = ["ReasonModel"]



from typing import Any, Dict, List, Union

from lion.integrations.perplexity_.api_endpoints.api_request import PerplexityRequest
from lion.integrations.perplexity_.api_endpoints.chat_completions.response.response_body import (
    PerplexityChatCompletionResponseBody,
)


def match_response(
    request_model: PerplexityRequest,
    response_body: dict[str, Any] | list[dict[str, Any]],
) -> PerplexityChatCompletionResponseBody | list[dict[str, Any]]:
    """Match the response body to the appropriate response model based on the endpoint."""

    if not response_body:
        return response_body

    if isinstance(response_body, list):
        return response_body

    if request_model.endpoint == "chat/completions":
        return PerplexityChatCompletionResponseBody(**response_body)

    return response_body



from typing import Literal

from pydantic import BaseModel, Field, model_validator

from ...data_models import AnthropicEndpointRequestBody
from .message_models import Message


class AnthropicMessageRequestBody(AnthropicEndpointRequestBody):
    model: str = Field(description="The model to use for generating the response")

    messages: list[Message] = Field(description="The messages to send to the model")

    max_tokens: int = Field(
        default=8000,  # Default to Anthropic's max model response tokens
        description="The maximum number of tokens to generate before stopping",
    )

    metadata: dict | None = Field(
        None, description="An object describing metadata about the request"
    )

    stop_sequences: list[str] | None = Field(
        None,
        description="Custom text sequences that will cause the model to stop generating",
    )

    stream: bool | None = Field(
        False, description="Whether to stream the response back token by token"
    )

    system: str | None = Field(
        None,
        description="System prompt that helps define the behavior of the assistant",
    )

    temperature: float | None = Field(
        None, ge=0, le=1, description="Amount of randomness injected into the response"
    )

    top_k: int | None = Field(
        None, description="Number of most likely tokens to consider at each step"
    )

    top_p: float | None = Field(
        None,
        ge=0,
        le=1,
        description="Total probability mass of tokens to consider at each step",
    )

    @model_validator(mode="after")
    def validate_messages(self):
        if not self.messages:
            raise ValueError("Messages list cannot be empty")

        # Check for alternating user/assistant messages
        for i in range(1, len(self.messages)):
            if self.messages[i].role == self.messages[i - 1].role:
                raise ValueError(
                    "Messages must alternate between user and assistant roles"
                )

        return self



"""
Copyright 2024 HaiyangLi

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""

from pydantic import BaseModel
from pydantic.fields import FieldInfo

from lion.core.typing import FieldModel, NewModelParams
from lion.protocols.operatives.operative import Operative

from .action import (
    ACTION_REQUESTS_FIELD,
    ACTION_REQUIRED_FIELD,
    ACTION_RESPONSES_FIELD,
    ActionRequestModel,
    ActionResponseModel,
)
from .reason import REASON_FIELD, ReasonModel


class StepModel(BaseModel):
    """Model representing a single operational step with optional reasoning and actions."""

    title: str
    description: str
    reason: ReasonModel | None = REASON_FIELD.field_info
    action_requests: list[ActionRequestModel] = ACTION_REQUESTS_FIELD.field_info
    action_required: bool = ACTION_REQUIRED_FIELD.field_info
    action_responses: list[ActionResponseModel] = ACTION_RESPONSES_FIELD.field_info


class Step:
    """Utility class providing methods to create and manage Operative instances for steps."""

    @staticmethod
    def request_operative(
        *,
        operative_name: str | None = None,
        reason: bool = False,
        actions: bool = False,
        request_params: NewModelParams | None = None,
        parameter_fields: dict[str, FieldInfo] | None = None,
        base_type: type[BaseModel] | None = None,
        field_models: list[FieldModel] | None = None,
        exclude_fields: list[str] | None = None,
        new_model_name: str | None = None,
        field_descriptions: dict[str, str] | None = None,
        inherit_base: bool = True,
        config_dict: dict | None = None,
        doc: str | None = None,
        frozen: bool = False,
    ) -> Operative:
        """Creates an Operative instance configured for request handling.

        Args:
            operative_name (str, optional): Name of the operative.
            reason (bool, optional): Whether to include reason field.
            actions (bool, optional): Whether to include action fields.
            request_params (NewModelParams, optional): Parameters for the new model.
            parameter_fields (dict[str, FieldInfo], optional): Parameter fields for the model.
            base_type (type[BaseModel], optional): Base type for the model.
            field_models (list[FieldModel], optional): List of field models.
            exclude_fields (list[str], optional): List of fields to exclude.
            new_model_name (str | None, optional): Name of the new model.
            field_descriptions (dict[str, str], optional): Descriptions for the fields.
            inherit_base (bool, optional): Whether to inherit base.
            config_dict (dict | None, optional): Configuration dictionary.
            doc (str | None, optional): Documentation string.
            frozen (bool, optional): Whether the model is frozen.

        Returns:
            Operative: The configured operative instance.
        """

        field_models = field_models or []
        exclude_fields = exclude_fields or []
        field_descriptions = field_descriptions or {}
        if reason:
            field_models.append(REASON_FIELD)
        if actions:
            field_models.extend(
                [
                    ACTION_REQUESTS_FIELD,
                    ACTION_REQUIRED_FIELD,
                ]
            )
        request_params = request_params or NewModelParams(
            parameter_fields=parameter_fields,
            base_type=base_type,
            field_models=field_models,
            exclude_fields=exclude_fields,
            name=new_model_name,
            field_descriptions=field_descriptions,
            inherit_base=inherit_base,
            config_dict=config_dict,
            doc=doc,
            frozen=frozen,
        )
        return Operative(name=operative_name, request_params=request_params)

    @staticmethod
    def respond_operative(
        *,
        operative: Operative,
        additional_data: dict | None = None,
        response_params: NewModelParams | None = None,
        field_models: list[FieldModel] | None = None,
        frozen_response: bool = False,
        response_config_dict: dict | None = None,
        response_doc: str | None = None,
        exclude_fields: list[str] | None = None,
    ) -> Operative:
        """Updates the operative with response parameters and data.

        Args:
            operative (Operative): The operative instance to update.
            additional_data (dict | None, optional): Additional data to include in the response.
            response_params (NewModelParams | None, optional): Parameters for the response model.
            field_models (list[FieldModel] | None, optional): List of field models.
            frozen_response (bool, optional): Whether the response model is frozen.
            response_config_dict (dict | None, optional): Configuration dictionary for the response.
            response_doc (str | None, optional): Documentation string for the response.
            exclude_fields (list[str] | None, optional): List of fields to exclude.

        Returns:
            Operative: The updated operative instance.
        """

        additional_data = additional_data or {}
        field_models = field_models or []
        operative = Step._create_response_type(
            operative=operative,
            response_params=response_params,
            field_models=field_models,
            frozen_response=frozen_response,
            response_config_dict=response_config_dict,
            response_doc=response_doc,
            exclude_fields=exclude_fields,
        )

        data = operative.response_model.model_dump()
        data.update(additional_data or {})
        operative.response_model = operative.response_type.model_validate(data)
        return operative

    @staticmethod
    def _create_response_type(
        operative: Operative,
        response_params: NewModelParams | None = None,
        response_validators: dict | None = None,
        frozen_response: bool = False,
        response_config_dict: dict | None = None,
        response_doc: str | None = None,
        field_models: list[FieldModel] | None = None,
        exclude_fields: list[str] | None = None,
    ) -> Operative:
        """Internal method to create a response type for the operative.

        Args:
            operative (Operative): The operative instance.
            response_params (NewModelParams | None, optional): Parameters for the response model.
            response_validators (dict | None, optional): Validators for the response model.
            frozen_response (bool, optional): Whether the response model is frozen.
            response_config_dict (dict | None, optional): Configuration dictionary for the response.
            response_doc (str | None, optional): Documentation string for the response.
            field_models (list[FieldModel] | None, optional): List of field models.
            exclude_fields (list[str] | None, optional): List of fields to exclude.

        Returns:
            Operative: The operative instance with updated response type.
        """

        field_models = field_models or []

        if (
            hasattr(operative.request_type, "action_required")
            and operative.response_model.action_required
        ):
            field_models.extend(
                [
                    ACTION_RESPONSES_FIELD,
                    ACTION_REQUIRED_FIELD,
                    ACTION_REQUESTS_FIELD,
                ]
            )
        if hasattr(operative.request_type, "reason"):
            field_models.extend([REASON_FIELD])

        exclude_fields = exclude_fields or []
        exclude_fields.extend(operative.request_params.exclude_fields)

        operative.create_response_type(
            response_params=response_params,
            field_models=field_models,
            exclude_fields=exclude_fields,
            doc=response_doc,
            config_dict=response_config_dict,
            frozen=frozen_response,
            validators=response_validators,
        )
        return operative



from pydantic import JsonValue

function_field_description = (
    "Specify the name of the function to execute. **Choose "
    "from the provided `tool_schemas`; do not invent function names.**"
    "Only provide function names if tool_schemas are provided. Otherwise, "
    "must leave blank or set to None."
)

arguments_field_description = (
    "Provide the arguments to pass to the function as a "
    "dictionary. **Use argument names and types as specified in the "
    "`tool_schemas`; do not invent argument names.**"
)

action_required_field_description = (
    "Specify whether the step requires actions to be "
    "performed. If **True**, the actions in `action_requests` "
    "must be performed. If **False**, the actions in "
    "`action_requests` are optional. If no tool_schemas"
    " are provided, this field is ignored."
)

action_requests_field_description = (
    "List of actions to be performed if `action_required` "
    "is **True**. Leave empty if no action is required. "
    "**When providing actions, you must choose from the "
    "provided `tool_schemas`. Do not invent function or "
    "argument names.**"
)

confidence_description = (
    "Provide an objective numeric confidence score between 0 and 1 (with 3 "
    "decimal places) indicating how likely you successfully achieved the task"
    " according to user expectation. Interpret the score as:\n"
    "- **1**: Very confident in a good job.\n"
    "- **0**: Not confident at all.\n"
    "- **[0.8, 1]**: You can continue the path of reasoning if needed.\n"
    "- **[0.5, 0.8)**: Recheck your reasoning and consider reverting to a "
    "previous, more confident reasoning path.\n"
    "- **[0, 0.5)**: Stop because the reasoning is starting to be off track."
)

instruction_field_description = (
    "Define the core task or instruction to be executed. Your instruction should:\n\n"
    "1. Be specific and actionable\n"
    "2. Clearly state the expected outcome\n"
    "3. Include any critical constraints or requirements\n\n"
    "**Guidelines for writing effective instructions:**\n"
    "- Start with a clear action verb (e.g., analyze, create, evaluate)\n"
    "- Specify the scope and boundaries of the task\n"
    "- Include success criteria when applicable\n"
    "- Break complex tasks into distinct steps\n\n"
    "**Examples:**\n"
    "- 'Analyze the provided sales data and identify top 3 performing products'\n"
    "- 'Generate a Python function that validates email addresses'\n"
    "- 'Create a data visualization showing monthly revenue trends'"
)

guidance_field_description = (
    "Provide strategic direction and constraints for task execution.\n\n"
    "**Key components to include:**\n"
    "1. Methodological preferences\n"
    "2. Quality standards and requirements\n"
    "3. Specific limitations or boundaries\n"
    "4. Performance expectations\n\n"
    "**Best practices:**\n"
    "- Be explicit about any assumptions that should be made\n"
    "- Specify preferred approaches or techniques\n"
    "- Detail any constraints on resources or methods\n"
    "- Include relevant standards or compliance requirements\n\n"
    "Leave as None if no specific guidance is needed beyond the instruction."
)

context_field_description = (
    "Supply essential background information and current state data required for "
    "task execution.\n\n"
    "**Include relevant details about:**\n"
    "1. Environmental conditions\n"
    "2. Historical context\n"
    "3. Related systems or processes\n"
    "4. Previous outcomes or decisions\n\n"
    "**Context should:**\n"
    "- Be directly relevant to the task\n"
    "- Provide necessary background without excess detail\n"
    "- Include any dependencies or prerequisites\n"
    "- Specify the current state of the system\n\n"
    "Set to None if no additional context is required."
)

reason_field_description = (
    "Control whether detailed reasoning should be included in the response.\n\n"
    "**When set to True:**\n"
    "- Must include a ReasonModel explaining decision rationale\n"
    "- Should detail key decision points\n"
    "- Must provide confidence scores for decisions\n"
    "- Should explain trade-offs considered\n\n"
    "**When set to False:**\n"
    "- Skip detailed reasoning\n"
    "- Focus on direct results\n"
    "- Omit confidence scoring\n\n"
    "Set to None to make reasoning optional based on context."
)

actions_field_description = (
    "Specify whether concrete actions should be taken as part of task execution.\n\n"
    "**When set to True:**\n"
    "- Must include appropriate ActionRequestModels\n"
    "- Actions should directly relate to task goals\n"
    "- Each action must be properly structured and validated\n"
    "- Actions must use available tool schemas\n\n"
    "**When set to False:**\n"
    "- No actions should be included\n"
    "- Focus on analysis and recommendations\n\n"
    "Set to None to make actions optional based on requirements."
)

# Example structures for each field to demonstrate proper formatting
instruction_examples: list[JsonValue] = [
    "Analyze the dataset 'sales_2023.csv' and identify revenue trends",
    "Create a Python function to process customer feedback data",
    {
        "task": "data_analysis",
        "target": "sales_performance",
        "scope": ["revenue", "growth", "seasonality"],
    },
]

guidance_examples: list[JsonValue] = [
    "Use statistical methods for trend analysis",
    "Optimize for readability and maintainability",
    {
        "methods": ["regression", "time_series"],
        "constraints": {"memory": "2GB", "time": "5min"},
    },
]

context_examples: list[JsonValue] = [
    "Previous analysis showed seasonal patterns",
    {
        "prior_results": {"accuracy": 0.95},
        "system_state": "production",
        "dependencies": ["numpy", "pandas"],
    },
]

"""Field description for InstructModel."""

instruct_model_description = (
    "Generate structured instructions for task execution.\n\n"
    "**Key Components:**\n"
    "1. Task Definition: Clear description of what needs to be accomplished\n"
    "2. Execution Parameters: How the task should be performed\n"
    "3. Success Criteria: What constitutes successful completion\n"
    "4. Scope and Boundaries: Limits and constraints of the task\n\n"
    "**Structure Guidelines:**\n"
    "- `instruction`: Core task or objective to accomplish\n"
    "- `guidance`: Parameters, preferences, and constraints\n"
    "- `context`: Relevant background and environmental information\n"
    "- `reason`: Whether to include reasoning in output\n"
    "- `actions`: Whether specific actions are required\n\n"
    "**Best Practices:**\n"
    "- Keep instructions clear and specific\n"
    "- Provide necessary but not excessive detail\n"
    "- Define measurable outcomes\n"
    "- Include relevant dependencies\n"
    "- Specify critical constraints\n\n"
    "**Common Issues to Avoid:**\n"
    "- Vague or ambiguous directives\n"
    "- Missing essential context\n"
    "- Undefined success criteria\n"
    "- Incomplete requirements\n"
    "- Conflicting parameters\n\n"
    "Structure instructions to enable successful task execution while maintaining "
    "appropriate flexibility for implementation details."
)

instruct_model_examples = [
    {
        "instruction": "Process the input data according to specified requirements",
        "guidance": {
            "requirements": ["validation", "transformation", "aggregation"],
            "output_format": "structured_report",
            "quality_metrics": ["accuracy", "completeness"],
        },
        "context": {
            "input_source": "data_stream",
            "domain": "general",
            "priority": "standard",
        },
        "reason": True,
        "actions": True,
    },
    {
        "instruction": "Evaluate system performance against baseline metrics",
        "guidance": {
            "evaluation_criteria": [
                "response_time",
                "resource_usage",
                "error_rate",
            ],
            "methodology": "standard",
            "reporting": "detailed",
        },
        "context": {
            "environment": "test",
            "criticality": "medium",
            "constraints": {"time": "bounded"},
        },
        "reason": True,
        "actions": False,
    },
]


__all__ = [
    "function_field_description",
    "arguments_field_description",
    "action_required_field_description",
    "action_requests_field_description",
    "confidence_description",
    "instruction_field_description",
    "guidance_field_description",
    "context_field_description",
    "reason_field_description",
    "actions_field_description",
    "instruction_examples",
    "guidance_examples",
    "context_examples",
    "instruct_model_description",
    "instruct_model_examples",
]



import re


def extract_code_block(
    str_to_parse: str,
    return_as_list: bool = False,
    languages: list[str] | None = None,
    categorize: bool = False,
) -> str | list[str] | dict[str, list[str]]:
    """
    Extract code blocks from a given string containing
    Markdown-formatted text.

    This function identifies code blocks enclosed by triple
    backticks (```) or
    tildes (~~~), extracts their content, and can filter
    them based on specified
    programming languages. It provides options to return
    the extracted code
    blocks as a single concatenated string, a list, or a
    dictionary categorized
    by language.

    Args:
        str_to_parse: The input string containing Markdown
        code blocks.
        return_as_list: If True, returns a list of code
        blocks; otherwise, returns
            them as a single concatenated string separated
            by two newlines.
        languages: A list of languages to filter the code
        blocks. If None,
            extracts code blocks of all languages.
        categorize: If True, returns a dictionary mapping
        languages to lists of
            code blocks.

    Returns:
        Depending on the parameters:
            - A concatenated string of code blocks.
            - A list of code blocks.
            - A dictionary mapping languages to lists of
            code blocks.
    """
    code_blocks = []
    code_dict = {}

    pattern = re.compile(
        r"""
        ^(?P<fence>```|~~~)[ \t]*     # Opening fence ``` or ~~~
        (?P<lang>[\w+-]*)[ \t]*\n     # Optional language identifier
        (?P<code>.*?)(?<=\n)          # Code content
        ^(?P=fence)[ \t]*$            # Closing fence matching the opening
        """,
        re.MULTILINE | re.DOTALL | re.VERBOSE,
    )

    for match in pattern.finditer(str_to_parse):
        lang = match.group("lang") or "plain"
        code = match.group("code")

        if languages is None or lang in languages:
            if categorize:
                code_dict.setdefault(lang, []).append(code)
            else:
                code_blocks.append(code)

    if categorize:
        return code_dict
    elif return_as_list:
        return code_blocks
    else:
        return "\n\n".join(code_blocks)



from pydantic import Field

from ..data_models import OllamaEndpointRequestBody


class OllamaDeleteModelRequestBody(OllamaEndpointRequestBody):
    name: str = Field(description="Name of the model to delete")



PROMPT = """Perform a brainstorm session. Fill in {num_instruct} Instruct for the appropriate next step, we will run them separately and concurrently with same external context, but you should supplement each idea with certain amount of uniqueness while adhering to the guidelines and standards of the project. The Instruct should be concisely informational. If you think a particular step requries further extension, you should mention it in the instruct"""



from .action import ActionRequestModel, ActionResponseModel
from .instruct import Instruct
from .operative import Operative
from .reason import ReasonModel
from .step import Step, StepModel

__all__: list[str] = [
    "Operative",
    "Step",
    "ActionRequestModel",
    "ActionResponseModel",
    "StepModel",
    "Instruct",
    "ReasonModel",
]
